from loguru import logger
from watchdog.observers.polling import PollingObserver as Observer

from tesseract_client.db_manager import DBManager
from tesseract_client.services import Services
from tesseract_client.monitoring import FileChangeHandler
from tesseract_client.api_manager import APIManager, API_URL
from tesseract_client.config import get_config, get_credentials, NoCredentialsError


def run():
    """Start the monitoring process"""
    try:
        username, password = get_credentials()
    except NoCredentialsError:
        logger.info("You are not logged in. Please run 'tesseract login'")
        return

    indexed_folder, db_path, api_url = get_config()

    api_urls = API_URL(base=api_url)

    with DBManager(db_path) as db_manager:
        api_manager = APIManager(
            username=username, password=password, api_urls=api_urls
        )
        chunk_size = api_manager.get_chunk_size()
        services = Services(
            api_manager=api_manager,
            db_manager=db_manager,
            indexed_folder=indexed_folder,
            chunk_size=chunk_size,
        )
        event_handler = FileChangeHandler(services)

        try:
            # Check for changes that were made while offline
            services.check_for_offline_changes()
            # Pull updates from the server
            services.pull()

            # Start monitoring the folder for changes
            observer = Observer()
            observer.schedule(event_handler, path=indexed_folder, recursive=True)
            observer.start()
            logger.info(f"Monitoring folder {indexed_folder} for updates...")

            try:
                while True:
                    pass
            except KeyboardInterrupt:
                observer.stop()
                logger.info("Monitoring stopped due to KeyboardInterrupt")

            observer.join()
        except Exception as e:
            logger.info(e)
            logger.error("Failed to start monitoring")
            return



import random
import string


def write_to_file(file_path: str, content: bytes) -> None:
    """Creates a sample file with the given content"""
    with open(file_path, 'wb') as file:
        file.write(content)


def append_to_file(file_path: str, content: bytes) -> None:
    """Appends the given content to the file"""
    with open(file_path, 'ab') as file:
        file.write(content)


def replace_in_file(file_path: str, start_index: int, end_index: int, content: bytes) -> None:
    """
    Replaces the content of the file from start_index (inclusive)
    to end_index (exclusive) with the given content.
    """
    with open(file_path, 'r+b') as file:
        text = file.read()
    text = text[:start_index] + content + text[end_index:]
    with open(file_path, 'wb') as file:
        file.write(text)


def generate_random_string(length: int) -> str:
    """Generates a random string of the given length"""
    return ''.join(random.choice(string.ascii_letters) for _ in range(length))



from loguru import logger

from tesseract_client.config import update_config


def config(index_path=None, db_path=None, api_url=None):
    """Configure the client"""
    if index_path or db_path or api_url:
        update_config(index_path, db_path, api_url)
        logger.info("Updated config file")
    else:
        logger.info("No changes made to config file")



import os
import math
import hashlib
from hypothesis import (
    given,
    settings,
    HealthCheck,
    strategies as st
)

from tests.utils import write_to_file, append_to_file, replace_in_file
from tesseract_client.file import File
from tesseract_client.chunk import Chunk, ChunkAction


def get_chunks(file_path: str, file: File, chunk_size: int) -> list[Chunk]:
    with open(file_path, 'rb') as f:
        return file.split_into_chunks(f, chunk_size)


def test_get_file_hash(tmpdir):
    file_content = b"Test Content"
    file_path = os.path.join(tmpdir, "test_file.txt")
    with open(file_path, 'wb') as f:
        f.write(file_content)

    expected_hash = hashlib.sha256(file_content).hexdigest()
    assert File.get_file_hash(file_path) == expected_hash


def test_get_relative_path():
    subdir = "subdir"
    file_path = os.path.join(subdir, "test_file.txt")
    assert File.get_relative_path(file_path, subdir) == "test_file.txt"
    assert File.get_relative_path(file_path, subdir + "/") == "test_file.txt"
    file_path = os.path.join(subdir, subdir, "test_file.txt")
    assert File.get_relative_path(file_path, subdir) == f"{subdir}/test_file.txt"
    subdir = '/home/user/subdir'
    file_path = os.path.join(subdir, "test_file.txt")
    assert File.get_relative_path(file_path, subdir) == "test_file.txt"


def test_from_local_file(tmpdir):
    file_path = os.path.join(tmpdir, "test_file.txt")
    write_to_file(file_path, b"Test Content")
    file = File.from_local_file(file_path, tmpdir)
    assert file.file_path == "test_file.txt"
    assert file.hash == File.get_file_hash(file_path)


@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(
    chunk_size=st.integers(min_value=1, max_value=1000),
    content=st.binary()
)
def test_split_into_chunks(tmpdir, chunk_size, content):
    file_path = os.path.join(tmpdir, "test_file.txt")
    write_to_file(file_path, content)
    file = File.from_local_file(file_path, tmpdir)
    with open(file_path, 'rb') as f:
        chunks = file.split_into_chunks(f, chunk_size)

    expected_chunk_count = math.ceil(len(content) / chunk_size)
    assert len(chunks) == expected_chunk_count

    # Split content into chunks of size chunk_size
    contents = [
        content[i:i + chunk_size] for i in range(0, len(content), chunk_size)
    ]

    order = 1
    for (chunk, data), expected_content in zip(chunks, contents):
        assert chunk.file_path == "test_file.txt"
        assert chunk.hash == hashlib.sha256(data).hexdigest()
        assert chunk.order == order
        assert data == expected_content
        order += 1


def test_get_updated_chunks_no_changes(tmpdir):
    file_path = os.path.join(tmpdir, "test_file.txt")
    content = b"test content"

    # Create file
    write_to_file(file_path, content)
    file = File.from_local_file(file_path, tmpdir)
    chunks = [chunk[0] for chunk in get_chunks(file_path, file, 1000)]

    # Update file
    with open(file_path, 'rb') as f:
        updated_chunks = file.get_updated_chunks(f, chunks, 1000)
    assert len(updated_chunks) == 0


def test_get_updated_chunks_new_chunks(tmpdir):
    file_path = os.path.join(tmpdir, "test_file.txt")
    content = b""

    # Create file
    write_to_file(file_path, content)
    file = File.from_local_file(file_path, tmpdir)
    chunks = [chunk[0] for chunk in get_chunks(file_path, file, 1000)]

    # Update file
    new_content = b"0123456789"
    append_to_file(file_path, new_content)
    with open(file_path, 'rb') as f:
        updated_chunks = file.get_updated_chunks(f, chunks, 10)

    assert len(updated_chunks) == 1
    assert updated_chunks[0][0][0].order == 1
    assert updated_chunks[0][0][0].hash == hashlib.sha256(new_content).hexdigest()
    assert updated_chunks[0][1] == ChunkAction.CREATE

    chunks = [chunk[0] for chunk in get_chunks(file_path, file, 1000)]

    # Append new chunk
    new_content = b"0123456789"
    append_to_file(file_path, content + new_content)
    with open(file_path, 'rb') as f:
        updated_chunks = file.get_updated_chunks(f, chunks, 10)

    assert len(updated_chunks) == 1
    assert updated_chunks[0][0][0].order == 2
    assert updated_chunks[0][0][0].hash == hashlib.sha256(new_content).hexdigest()
    assert updated_chunks[0][1] == ChunkAction.CREATE


def test_get_updated_chunks_deleted_chunks(tmpdir):
    file_path = os.path.join(tmpdir, "test_file.txt")
    content = b"0123456789"

    # Create file
    write_to_file(file_path, content)
    file = File.from_local_file(file_path, tmpdir)
    chunks = [chunk[0] for chunk in get_chunks(file_path, file, 10)]

    # Delete chunk
    write_to_file(file_path, b"")
    with open(file_path, 'rb') as f:
        updated_chunks = file.get_updated_chunks(f, chunks, 10)

    assert len(updated_chunks) == 1
    assert updated_chunks[0][0][0].order == 1
    assert updated_chunks[0][0][1] is None
    assert updated_chunks[0][1] == ChunkAction.DELETE


def test_get_updated_chunks_updated_chunks(tmpdir):
    file_path = os.path.join(tmpdir, "test_file.txt")
    content = b"0123456789"

    # Create file
    write_to_file(file_path, content)
    file = File(file_path, str(tmpdir))
    chunks = [chunk[0] for chunk in get_chunks(file_path, file, 10)]

    # Update chunk
    new_content = b"9876543210"
    replace_in_file(file_path, 0, 10, new_content)
    with open(file_path, 'rb') as f:
        updated_chunks = file.get_updated_chunks(f, chunks, 10)

    assert len(updated_chunks) == 1
    assert updated_chunks[0][0][0].order == 1
    assert updated_chunks[0][0][0].hash == hashlib.sha256(new_content).hexdigest()
    assert updated_chunks[0][1] == ChunkAction.UPDATE


def test_get_updated_chunks_multiple_changes(tmpdir):
    file_path = os.path.join(tmpdir, "test_file.txt")
    content = b"0123456789"

    # Create file
    write_to_file(file_path, content)
    file = File(file_path, str(tmpdir))
    chunks = [chunk[0] for chunk in get_chunks(file_path, file, 10)]

    # Update chunk
    new_content = b"9876543210"
    write_to_file(file_path, new_content)
    with open(file_path, 'rb') as f:
        updated_chunks = file.get_updated_chunks(f, chunks, 10)

    assert len(updated_chunks) == 1
    assert updated_chunks[0][0][0].order == 1
    assert updated_chunks[0][0][0].hash == hashlib.sha256(new_content).hexdigest()
    assert updated_chunks[0][1] == ChunkAction.UPDATE

    chunks = [chunk[0] for chunk in get_chunks(file_path, file, 10)]

    # Delete chunk
    write_to_file(file_path, b"")
    with open(file_path, 'rb') as f:
        updated_chunks = file.get_updated_chunks(f, chunks, 10)

    assert len(updated_chunks) == 1
    assert updated_chunks[0][0][0].order == 1
    assert updated_chunks[0][0][1] is None
    assert updated_chunks[0][1] == ChunkAction.DELETE

    chunks = [chunk[0] for chunk in get_chunks(file_path, file, 10)]

    # Append new chunk
    new_content = b"0123456789"
    write_to_file(file_path, new_content)
    with open(file_path, 'rb') as f:
        updated_chunks = file.get_updated_chunks(f, chunks, 10)

    assert len(updated_chunks) == 1
    assert updated_chunks[0][0][0].order == 1
    assert updated_chunks[0][0][0].hash == hashlib.sha256(new_content).hexdigest()
    assert updated_chunks[0][1] == ChunkAction.CREATE


def test_get_updated_chunks_updated_in_middle(tmpdir):
    file_path = os.path.join(tmpdir, "test_file.txt")
    content = b"0123456789"

    # Create file
    write_to_file(file_path, content)
    file = File(file_path, str(tmpdir))
    chunks = [chunk[0] for chunk in get_chunks(file_path, file, 2)]

    # Update chunk with the same size
    replace_in_file(file_path, 2, 4, b"21")
    with open(file_path, 'rb') as f:
        updated_chunks = file.get_updated_chunks(f, chunks, 2)
    assert len(updated_chunks) == 1
    assert updated_chunks[0][0][0].order == 2
    assert updated_chunks[0][0][0].hash == hashlib.sha256(b"21").hexdigest()
    assert updated_chunks[0][1] == ChunkAction.UPDATE

    chunks = [chunk[0] for chunk in get_chunks(file_path, file, 2)]

    # Update chunk with a different size
    replace_in_file(file_path, 2, 4, b"3210")
    with open(file_path, 'rb') as f:
        updated_chunks = file.get_updated_chunks(f, chunks, 2)
    # 01 21 45 67 89 -> 01 32 10 45 67 89
    assert len(updated_chunks) == 5
    print(updated_chunks)
    assert updated_chunks[0][0][0].order == 6
    assert updated_chunks[0][0][1] == b"89"
    assert updated_chunks[0][1] == ChunkAction.CREATE
    assert updated_chunks[1][0][0].order == 2
    assert updated_chunks[1][0][1] == b"32"
    assert updated_chunks[1][1] == ChunkAction.UPDATE
    assert updated_chunks[2][0][0].order == 3
    assert updated_chunks[2][0][1] == b"10"
    assert updated_chunks[2][1] == ChunkAction.UPDATE
    assert updated_chunks[3][0][0].order == 4
    assert updated_chunks[3][0][1] == b"45"
    assert updated_chunks[3][1] == ChunkAction.UPDATE
    assert updated_chunks[4][0][0].order == 5
    assert updated_chunks[4][0][1] == b"67"
    assert updated_chunks[4][1] == ChunkAction.UPDATE



import pytest
import tempfile
import os
import sqlite3

from tesseract_client.file import File
from tesseract_client.chunk import Chunk
from tesseract_client.db_manager import DBManager


@pytest.fixture
def db_manager() -> DBManager:
    """Creates a temporary database and returns a DBManager instance"""
    db_fd, db_path = tempfile.mkstemp()
    with DBManager(db_path) as db:
        yield db
    os.close(db_fd)
    os.unlink(db_path)


def test_create_file(db_manager: DBManager):
    file = File(file_path='/path/to/file.txt', hash='file_hash')
    db_manager.create_file(file)

    retrieved_file = db_manager.get_file_by_path('/path/to/file.txt')
    assert retrieved_file.file_path == '/path/to/file.txt'
    assert retrieved_file.hash == 'file_hash'


def test_create_chunk(db_manager: DBManager):
    chunk = Chunk(file_path='/path/to/file.txt', order=1, hash='chunk_hash')
    db_manager.create_chunk(chunk)

    chunks = db_manager.get_chunks('/path/to/file.txt')
    assert len(chunks) == 1
    retrieved_chunk = chunks[0]
    assert retrieved_chunk.file_path == '/path/to/file.txt'
    assert retrieved_chunk.order == 1
    assert retrieved_chunk.hash == 'chunk_hash'


def test_update_file(db_manager: DBManager):
    file = File(file_path='/path/to/file.txt', hash='file_hash')
    db_manager.create_file(file)

    updated_file = File(file_path='/path/to/file.txt', hash='new_file_hash')
    db_manager.update_file(updated_file)

    retrieved_file = db_manager.get_file_by_path('/path/to/file.txt')
    assert retrieved_file.file_path == '/path/to/file.txt'
    assert retrieved_file.hash == 'new_file_hash'


def test_update_chunk(db_manager: DBManager):
    chunk = Chunk(file_path='/path/to/file.txt', order=1, hash='chunk_hash')
    db_manager.create_chunk(chunk)

    updated_chunk = Chunk(
        file_path='/path/to/file.txt',
        order=1,
        hash='new_chunk_hash'
    )
    db_manager.update_chunk(updated_chunk)

    chunks = db_manager.get_chunks('/path/to/file.txt')
    assert len(chunks) == 1
    retrieved_chunk = chunks[0]
    assert retrieved_chunk.file_path == '/path/to/file.txt'
    assert retrieved_chunk.order == 1
    assert retrieved_chunk.hash == 'new_chunk_hash'


def test_delete_file(db_manager: DBManager):
    file = File(file_path='/path/to/file.txt', hash='file_hash')
    db_manager.create_file(file)

    db_manager.delete_file('/path/to/file.txt')

    retrieved_file = db_manager.get_file_by_path('/path/to/file.txt')
    assert retrieved_file is None


def test_delete_chunk(db_manager: DBManager):
    chunk = Chunk(file_path='/path/to/file.txt', order=1, hash='chunk_hash')
    db_manager.create_chunk(chunk)

    db_manager.delete_chunk(chunk.file_path, chunk.order)

    chunks = db_manager.get_chunks('/path/to/file.txt')
    assert len(chunks) == 0


def test_get_files(db_manager: DBManager):
    file1 = File(file_path='/path/to/file1.txt', hash='file1_hash')
    file2 = File(file_path='/path/to/file2.txt', hash='file2_hash')
    db_manager.create_file(file1)
    db_manager.create_file(file2)

    files = db_manager.get_files()
    assert len(files) == 2
    assert files[0].file_path == '/path/to/file1.txt'
    assert files[0].hash == 'file1_hash'
    assert files[1].file_path == '/path/to/file2.txt'
    assert files[1].hash == 'file2_hash'


def test_get_chunks(db_manager: DBManager):
    chunk1 = Chunk(file_path='/path/to/file.txt', order=1, hash='chunk1_hash')
    chunk2 = Chunk(file_path='/path/to/file.txt', order=2, hash='chunk2_hash')
    db_manager.create_chunk(chunk1)
    db_manager.create_chunk(chunk2)

    chunks = db_manager.get_chunks('/path/to/file.txt')
    assert len(chunks) == 2
    assert chunks[0].file_path == '/path/to/file.txt'
    assert chunks[0].order == 1
    assert chunks[0].hash == 'chunk1_hash'
    assert chunks[1].file_path == '/path/to/file.txt'
    assert chunks[1].order == 2
    assert chunks[1].hash == 'chunk2_hash'


def test_get_file_by_path(db_manager: DBManager):
    file = File(file_path='/path/to/file.txt', hash='file_hash')
    db_manager.create_file(file)

    retrieved_file = db_manager.get_file_by_path('/path/to/file.txt')
    assert retrieved_file.file_path == '/path/to/file.txt'
    assert retrieved_file.hash == 'file_hash'


def test_get_file_by_path_returns_none_if_not_found(db_manager: DBManager):
    retrieved_file = db_manager.get_file_by_path('/path/to/file.txt')
    assert retrieved_file is None


def test_file_name_unique(db_manager: DBManager):
    file1 = File(file_path='/path/to/file.txt', hash='file_hash')
    file2 = File(file_path='/path/to/file.txt', hash='file_hash')
    db_manager.create_file(file1)
    with pytest.raises(sqlite3.IntegrityError):
        db_manager.create_file(file2)


def test_chunk_unique(db_manager: DBManager):
    chunk1 = Chunk(file_path='/path/to/file.txt', order=1, hash='chunk_hash')
    chunk2 = Chunk(file_path='/path/to/file.txt', order=1, hash='chunk_hash1')
    db_manager.create_chunk(chunk1)
    with pytest.raises(sqlite3.IntegrityError):
        db_manager.create_chunk(chunk2)



from loguru import logger

from tesseract_client.config import delete_credentials, clean_up, NoCredentialsError


def logout():
    """Deletes the credentials from the config file and keyring and cleans up the file system"""
    try:
        clean_up()
    except FileNotFoundError:
        pass

    try:
        delete_credentials()
    except NoCredentialsError:
        logger.info("You are not logged in")
        return

    logger.info("Logged out")



import pytest
from pathlib import Path

from tesseract_client.config import (
    create_config_if_not_exists,
    update_config_file,
    update_config,
    get_config,
    store_credentials,
    get_credentials,
    delete_credentials,
    clean_up,
    NoCredentialsError
)
from tesseract_client.file import File


def test_create_config_if_not_exists(monkeypatch, tmpdir):
    mock_config_path = tmpdir.join('config').join('config.ini')
    monkeypatch.setattr('tesseract_client.config.CONFIG_PATH', str(mock_config_path))

    index_path = str(tmpdir.join('index'))
    db_path = str(tmpdir.join('db').join('db.sqlite'))
    api_url = 'http://localhost:8000'

    create_config_if_not_exists(index_path, db_path, api_url)

    with open(str(mock_config_path), 'r') as configfile:
        assert configfile.read().strip() == (
            '[CONFIG]\n'
            f'index_path = {index_path}\n'
            f'db_path = {db_path}\n'
            f'api_url = {api_url}'
        )


def test_update_config_file(monkeypatch, tmpdir):
    mock_config_path = tmpdir.join('config').join('config.ini')
    monkeypatch.setattr('tesseract_client.config.CONFIG_PATH', str(mock_config_path))

    create_config_if_not_exists('index_path', 'db_path', 'api_url')

    update_config_file(
        index_path='new_index_path',
        db_path='new_db_path',
        api_url='new_api_url'
    )

    with open(str(mock_config_path), 'r') as configfile:
        assert configfile.read().strip() == (
            '[CONFIG]\n'
            'index_path = new_index_path\n'
            'db_path = new_db_path\n'
            'api_url = new_api_url'
        )


def test_update_config(monkeypatch, tmpdir):
    mock_config_path = tmpdir.join('config').join('config.ini')
    monkeypatch.setattr('tesseract_client.config.CONFIG_PATH', str(mock_config_path))

    index_path = str(tmpdir.join('index'))
    db_path = str(tmpdir.join('db.sqlite'))
    api_url = 'http://localhost:8000'

    create_config_if_not_exists(index_path, db_path, api_url)

    File.create_folder_path_if_not_exists(index_path, False)
    File.create_folder_path_if_not_exists(db_path, True)
    Path(db_path).touch()

    assert Path(index_path).exists()
    assert Path(db_path).exists()

    new_index_path = str(tmpdir.join('new_index'))
    new_db_path = str(tmpdir.join('new_db.sqlite'))
    new_api_url = 'http://localhost:8001'

    update_config(
        index_path=new_index_path,
        db_path=new_db_path,
        api_url=new_api_url
    )

    assert Path(index_path).exists() is False
    assert Path(db_path).exists() is False

    assert Path(new_index_path).exists()
    assert Path(new_db_path).exists()

    with open(str(mock_config_path), 'r') as configfile:
        assert configfile.read().strip() == (
            '[CONFIG]\n'
            f'index_path = {new_index_path}\n'
            f'db_path = {new_db_path}\n'
            f'api_url = {new_api_url}'
        )


def test_get_config(monkeypatch, tmpdir):
    mock_config_path = tmpdir.join('config').join('config.ini')
    monkeypatch.setattr('tesseract_client.config.CONFIG_PATH', str(mock_config_path))

    create_config_if_not_exists('index_path', 'db_path', 'api_url')

    assert get_config() == (
        'index_path',
        'db_path',
        'api_url'
    )


def test_store_credentials(monkeypatch, tmpdir, mocker):
    mock_keyring = mocker.patch('keyring.set_password')

    mock_config_path = tmpdir.join("config").join('config.ini')
    monkeypatch.setattr('tesseract_client.config.CONFIG_PATH', str(mock_config_path))

    create_config_if_not_exists('index_path', 'db_path', 'api_url')

    store_credentials('username', 'password')

    with open(str(mock_config_path), 'r') as configfile:
        config_content = configfile.read().strip()
        expected_content = (
            '[CONFIG]\n'
            'index_path = index_path\n'
            'db_path = db_path\n'
            'api_url = api_url\n\n'
            '[CREDENTIALS]\n'
            'username = username'
        )
        assert config_content == expected_content

    mock_keyring.assert_called_once_with('tesseract', 'username', 'password')


def test_get_credentials(monkeypatch, tmpdir, mocker):
    mock_keyring_set = mocker.patch('keyring.set_password')
    mock_keyring_get = mocker.patch('keyring.get_password')

    mock_config_path = tmpdir.join('config').join('config.ini')
    monkeypatch.setattr('tesseract_client.config.CONFIG_PATH', str(mock_config_path))

    create_config_if_not_exists('index_path', 'db_path', 'api_url')
    store_credentials('username', 'password')
    mock_keyring_get.return_value = 'password'

    assert get_credentials() == ('username', 'password')
    mock_keyring_set.assert_called_once_with('tesseract', 'username', 'password')
    mock_keyring_get.assert_called_once_with('tesseract', 'username')


def test_delete_credentials(monkeypatch, tmpdir, mocker):
    mocker.patch('keyring.set_password')
    mocker.patch('keyring.get_password')
    mocker.patch('keyring.delete_password')

    mock_config_path = tmpdir.join('config').join('config.ini')
    monkeypatch.setattr('tesseract_client.config.CONFIG_PATH', str(mock_config_path))

    create_config_if_not_exists('index_path', 'db_path', 'api_url')
    store_credentials('username', 'password')

    delete_credentials()

    with open(str(mock_config_path), 'r') as configfile:
        assert configfile.read().strip() == (
            '[CONFIG]\n'
            'index_path = index_path\n'
            'db_path = db_path\n'
            'api_url = api_url'
        )

    with pytest.raises(NoCredentialsError):
        get_credentials()


def test_clean_up(monkeypatch, tmpdir):
    mock_config_path = tmpdir.join('config').join('config.ini')
    mock_db_path = tmpdir.join('db').join('db.sqlite')
    mock_index_path = tmpdir.join('index')
    monkeypatch.setattr('tesseract_client.config.CONFIG_PATH', str(mock_config_path))

    create_config_if_not_exists('index_path', 'db_path', 'api_url')
    update_config_file(
        index_path=str(mock_index_path),
        db_path=str(mock_db_path)
    )

    # Create mock files
    Path(mock_db_path).parent.mkdir(parents=True)
    mock_db_path.write('')
    Path(mock_index_path).mkdir()

    assert Path(mock_db_path).exists()
    assert Path(mock_index_path).exists()

    clean_up()

    assert not Path(mock_db_path).exists()
    assert not Path(mock_index_path).exists()



import getpass
from loguru import logger

from tesseract_client.config import store_credentials, get_config, clean_up
from tesseract_client.api_manager import APIManager, API_URL


def login(username: str = None, password: str = None):
    """Validate and store the username and password in the config file and keyring"""
    if username is None:
        username = input("Username: ")
    if password is None:
        password = getpass.getpass("Password: ")

    _, _, api_url = get_config()
    api_manager = APIManager(
        username=username, password=password, api_urls=API_URL(api_url)
    )
    try:
        api_manager.login()
        logger.info(f"Logged in as {username}")
    except Exception as e:
        logger.info(e)
        return

    try:
        clean_up()
    except FileNotFoundError:
        pass

    store_credentials(username, password)



import functools
import requests
from loguru import logger
from dataclasses import dataclass

from tesseract_client.file import Chunk, File


class InvalidCredentials(Exception):
    pass


class ServerError(Exception):
    def __init__(self, msg="Server error"):
        super().__init__(msg)


class NotFoundError(Exception):
    def __init__(self):
        super().__init__("Not found")


@dataclass
class API_URL:
    base: str

    @property
    def signup(self) -> str:
        return self.base + "/auth/signup"

    @property
    def login(self) -> str:
        return self.base + "/auth/signin"

    @property
    def get_chunk_size(self) -> str:
        return self.base + "/chunk/size"

    @property
    def upload_file(self) -> str:
        return self.base + "/file"

    @property
    def delete_file(self) -> str:
        return self.base + "/file"

    @property
    def get_file(self) -> str:
        return self.base + "/file"

    @property
    def get_files(self) -> str:
        return self.base + "/file/all"

    @property
    def get_file_chunks(self) -> str:
        return self.base + "/file/chunks"

    @property
    def upload_chunk(self) -> str:
        return self.base + "/chunk"

    @property
    def delete_chunk(self) -> str:
        return self.base + "/chunk"

    @property
    def get_chunk(self) -> str:
        return self.base + "/chunk"

    @property
    def get_chunk_url(self) -> str:
        return self.base + "/chunk/download"


class APIManager:
    """Handles communication with the server"""

    def __init__(self, username: str, password: str, api_urls: API_URL):
        self.access_token = None
        self.username = username
        self.password = password
        self.api_urls = api_urls

    def signup(self):
        """Signs up to the server"""
        response = requests.post(
            self.api_urls.signup,
            data={"username": self.username, "password": self.password},
        )
        if response.status_code == 201:
            return response.json()
        elif response.status_code == 400:
            raise InvalidCredentials("Username already exists")
        elif response.status_code == 500:
            raise ServerError()

    def login(self):
        """Logs in to the server and stores the access token"""
        try:
            response = requests.post(
                self.api_urls.login,
                data={"username": self.username, "password": self.password},
            )
        except requests.exceptions.ConnectionError:
            raise ServerError()
        if response.status_code == 200:
            self.access_token = response.json()["access_token"]
            logger.debug("Obtained access token")
        elif response.status_code == 401:
            raise InvalidCredentials("Invalid credentials")
        elif response.status_code == 404:
            raise NotFoundError()
        else:
            raise ServerError()

    def renew_token(func):
        """Decorator that renews the access token if it has expired"""

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except InvalidCredentials:
                logger.debug("Access token expired. Renewing...")
                self.login()
                return func(self, *args, **kwargs)

        return wrapper

    @renew_token
    def _make_request(
        self, method: requests.Request, url: str, status_code: int, **kwargs
    ) -> requests.Response:
        """Makes a request to the server"""
        try:
            response = method(
                url, headers={"Authorization": f"Bearer {self.access_token}"}, **kwargs
            )
        except requests.exceptions.ConnectionError:
            raise ServerError("Could not connect to server")
        if response.status_code == status_code:
            return response
        if response.status_code == 401:
            raise InvalidCredentials("Invalid access token")
        elif response.status_code == 404:
            raise NotFoundError()
        else:
            raise ServerError()

    def get_chunk_size(self) -> int:
        """Gets the chunk size from the server"""
        return self._make_request(
            requests.get, self.api_urls.get_chunk_size, status_code=200
        ).json()

    def upload_file(self, file: File) -> None:
        """Uploads a file to the server"""
        self._make_request(
            requests.post,
            self.api_urls.upload_file,
            status_code=201,
            params={"path": file.file_path, "hash": file.hash},
        )

    def update_file(self, file: File) -> None:
        """Updates a file on the server"""
        self._make_request(
            requests.put,
            self.api_urls.upload_file,
            status_code=200,
            params={"path": file.file_path, "hash": file.hash},
        )

    def delete_file(self, file_path: str) -> None:
        """Deletes a file from the server"""
        self._make_request(
            requests.delete,
            self.api_urls.delete_file,
            status_code=204,
            params={"path": file_path},
        )

    def get_file(self, file_path: str) -> File:
        """Gets the metadata of a file from the server"""
        file = self._make_request(
            requests.get,
            self.api_urls.get_file,
            status_code=200,
            params={"path": file_path},
        ).json()
        return File(file["path"], file["hash"])

    def get_files(self) -> list[File]:
        """Gets all files from the server"""
        files = self._make_request(
            requests.get, self.api_urls.get_files, status_code=200
        ).json()
        return [File(file["path"], file["hash"]) for file in files]

    def get_file_chunks(self, file_path: str) -> list[Chunk]:
        """Gets all chunks of a file from the server"""
        chunks = self._make_request(
            requests.get,
            self.api_urls.get_file_chunks,
            status_code=200,
            params={"path": file_path},
        ).json()
        return [
            Chunk(chunk["file_path"], chunk["order"], chunk["hash"]) for chunk in chunks
        ]

    def upload_chunk(self, chunk: Chunk, data: bytes) -> None:
        """Uploads a chunk to the server"""
        self._make_request(
            requests.post,
            self.api_urls.upload_chunk,
            status_code=201,
            params={
                "file_path": chunk.file_path,
                "order": chunk.order,
                "hash": chunk.hash,
            },
            files={
                "file": data,
            },
        )

    def update_chunk(self, chunk: Chunk, data: bytes) -> None:
        """Updates a chunk on the server"""
        self._make_request(
            requests.put,
            self.api_urls.upload_chunk,
            status_code=200,
            params={
                "file_path": chunk.file_path,
                "order": chunk.order,
                "hash": chunk.hash,
            },
            files={"file": data},
        )

    def delete_chunk(self, file_path: str, order: int) -> None:
        """Deletes a chunk from the server"""
        self._make_request(
            requests.delete,
            self.api_urls.delete_chunk,
            status_code=204,
            params={"file_path": file_path, "order": order},
        )

    def get_chunk_meta(self, file_path: str, order: int) -> Chunk:
        """Gets the metadata of a chunk from the server"""
        chunk = self._make_request(
            requests.get,
            self.api_urls.get_chunk,
            status_code=200,
            params={"file_path": file_path, "order": order},
        ).json()
        return Chunk(chunk["file_path"], chunk["order"], chunk["hash"])

    def download_chunk(self, file_path: str, order: int) -> bytes:
        """Downloads a chunk from the server"""
        url = self._make_request(
            requests.get,
            self.api_urls.get_chunk_url,
            status_code=200,
            params={"file_path": file_path, "order": order},
        ).json()["url"]
        return requests.get(url).content


if __name__ == "__main__":
    api_manager = APIManager("test", "test", API_URL("http://localhost:8000"))
    api_manager.login()
    print(api_manager.access_token)



import sys
import argparse


from importlib.metadata import version

from loguru import logger

from tesseract_client import DEFAULT_INDEX_PATH, DEFAULT_DB_PATH, DEFAULT_API_URL
from tesseract_client.file import File
from tesseract_client.config import create_config_if_not_exists, get_config
from tesseract_client.cli.login import login
from tesseract_client.cli.run import run
from tesseract_client.cli.config import config
from tesseract_client.cli.signup import signup
from tesseract_client.cli.logout import logout
from tesseract_client.cli.pull import pull


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-V", "--version", action="version", version=f"Tesseract {version(__package__)}"
    )

    parser.add_argument(
        "--log-level",
        help="Log level",
        type=str.upper,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    subparsers = parser.add_subparsers(dest="command")

    # Signup
    signup_parser = subparsers.add_parser("signup", help="Create a new account")
    signup_parser.add_argument(
        "--username", "-u", help="Username to use for authentication", type=str
    )
    signup_parser.add_argument(
        "--password", "-p", help="Password to use for authentication", type=str
    )

    # Login
    login_parser = subparsers.add_parser("login", help="Login to the API")
    login_parser.add_argument(
        "--username", "-u", help="Username to use for authentication", type=str
    )
    login_parser.add_argument(
        "--password", "-p", help="Password to use for authentication", type=str
    )

    # Logout
    subparsers.add_parser("logout", help="Logout of the API")

    # Run
    subparsers.add_parser("run", help="Start monitoring")

    # Config
    config_parser = subparsers.add_parser("config", help="Configure Tesseract")
    config_parser.add_argument("--path", help="Folder to index", type=str, default=None)
    config_parser.add_argument(
        "--db", help="Path to database file", type=str, default=None
    )
    config_parser.add_argument(
        "--api_url", help="URL of the API", type=str, default=None
    )

    # Pull
    subparsers.add_parser("pull", help="Pull files from the server")

    args = parser.parse_args()

    logger.remove()
    logger.add(
        sys.stdout,
        level=args.log_level,
        format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level}</level> | <level>{message}</level>",
    )

    create_config_if_not_exists(DEFAULT_INDEX_PATH, DEFAULT_DB_PATH, DEFAULT_API_URL)
    indexed_folder, db_path, _ = get_config()

    File.create_folder_path_if_not_exists(indexed_folder, False)
    File.create_folder_path_if_not_exists(db_path, True)

    if args.command == "signup":
        signup(args.username, args.password)
    elif args.command == "login":
        login(args.username, args.password)
    elif args.command == "logout":
        logout()
    elif args.command == "run":
        run()
    elif args.command == "config":
        config(args.path, args.db, args.api_url)
    elif args.command == "pull":
        pull()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import os
import hashlib
from pathlib import Path
from typing import TextIO
from loguru import logger

from tesseract_client.chunk import Chunk, ChunkAction


class File:
    """Represents a file on the client machine."""

    def __init__(self, file_path: str, hash: str):
        self.file_path = file_path
        self.hash = hash
        logger.debug(f"Initialized File object for {file_path}")

    def split_into_chunks(
        self, fd: TextIO, chunk_size: int
    ) -> list[tuple[Chunk, bytes]]:
        """
        Returns a list of Chunk objects along with their data
        for the given file.
        """
        chunks = []
        chunk_num = 1
        while True:
            data = fd.read(chunk_size)
            if not data:
                break

            chunk_hash = hashlib.sha256(data).hexdigest()
            chunk = Chunk(file_path=self.file_path, order=chunk_num, hash=chunk_hash)
            chunks.append((chunk, data))
            chunk_num += 1
        logger.debug(f"Split {self.file_path} into {len(chunks)} chunks")
        return chunks

    def get_updated_chunks(
        self, fd: TextIO, indexed_chunks: list[Chunk], chunk_size: int
    ) -> list[tuple[Chunk, bytes], ChunkAction]:
        """
        Returns a list of chunks that have been updated since the last time
        the file was indexed.
        """
        chunks = self.split_into_chunks(fd, chunk_size)
        data = [chunk[1] for chunk in chunks]
        chunks = [chunk[0] for chunk in chunks]
        updated_chunks = Chunk.get_updated_chunks(indexed_chunks, chunks)
        result = []
        for chunk, action in updated_chunks:
            if action == ChunkAction.DELETE:
                result.append(((chunk, None), action))
            else:
                result.append(((chunk, data[chunk.order - 1]), action))
        return result

    @staticmethod
    def get_relative_path(file_path, root_folder) -> str:
        """Returns relative file path to the root folder"""
        relative_path = os.path.relpath(file_path, root_folder)
        logger.debug(f"Relative path for {file_path} is {relative_path}")
        return relative_path

    @staticmethod
    def get_file_hash(file_path: str) -> str:
        """Returns the hash of a file"""
        with open(file_path, "rb") as file:
            file_hash = hashlib.sha256(file.read()).hexdigest()
        logger.debug(f"File hash for {file_path} is {file_hash}")
        return file_hash

    @staticmethod
    def create_folder_path_if_not_exists(path: str, is_file: bool) -> None:
        """Create the full path to the folder if it doesn't exist"""
        path = Path(path)
        if is_file:
            path = path.parent
        path.mkdir(parents=True, exist_ok=True)
        logger.debug(f"Created path to: {path}")

    @classmethod
    def from_local_file(cls, file_path: str, root_folder: str):
        """Returns a File object from an absolute file path"""
        logger.debug(f"Created File object from local file {file_path}")
        return cls(
            cls.get_relative_path(file_path, root_folder), cls.get_file_hash(file_path)
        )

    def __repr__(self):
        return f"File({self.file_path}, {self.hash})"

    def __eq__(self, other):
        return self.file_path == other.file_path and self.hash == other.hash



from watchdog.events import FileSystemEventHandler
from loguru import logger

from tesseract_client.services import Services


class FileChangeHandler(FileSystemEventHandler):
    """Handles changes in the monitored folder"""

    def __init__(self, services: Services):
        super().__init__()
        self.services = services
        logger.debug("FileChangeHandler object initialized")

    def on_created(self, event):
        if not event.is_directory:
            self.services.create_file(event.src_path)

    def on_deleted(self, event):
        if not event.is_directory:
            self.services.delete_file(event.src_path)

    def on_modified(self, event):
        if not event.is_directory and event.event_type == "modified":
            self.services.update_file(event.src_path)



import getpass
from loguru import logger

from tesseract_client.api_manager import APIManager, API_URL
from tesseract_client.config import get_config


def signup(username: str = None, password: str = None):
    """Sign the user up for a new account"""
    if username is None:
        username = input("Username: ")
    if password is None:
        password = getpass.getpass("Password: ")

    _, _, api_url = get_config()
    api_manager = APIManager(
        username=username, password=password, api_urls=API_URL(api_url)
    )
    try:
        api_manager.signup()
    except Exception as e:
        logger.info(e)
        return
    logger.info(f"Created account for {username}")



import os
import math
import shutil
import hashlib
import tempfile
import pytest

from unittest.mock import Mock
from tesseract_client.api_manager import APIManager
from tesseract_client.db_manager import DBManager
from tesseract_client.services import Services
from tesseract_client.chunk import Chunk
from tests.utils import write_to_file, append_to_file, generate_random_string


INDEXED_FOLDER = "./indexed_folder"
CHUNK_SIZE = 10


@pytest.fixture
def services() -> Services:
    """Creates a temporary database and returns a Services instance"""
    db_fd, db_path = tempfile.mkstemp()
    os.mkdir(INDEXED_FOLDER)
    with DBManager(db_path) as db:
        api_manager = Mock(APIManager)
        yield Services(api_manager, db, INDEXED_FOLDER, CHUNK_SIZE)
    os.close(db_fd)
    os.unlink(db_path)
    shutil.rmtree(INDEXED_FOLDER)


def test_create_chunk(services: Services):
    chunk = Chunk("test_file.txt", 1, "test_hash")

    data = b"Test Content"
    services.create_chunk(chunk, data)

    indexed_chunks = services.db_manager.get_chunks("test_file.txt")
    assert len(indexed_chunks) == 1
    indexed_chunk = indexed_chunks[0]
    assert indexed_chunk.file_path == chunk.file_path
    assert indexed_chunk.order == chunk.order
    assert indexed_chunk.hash == chunk.hash
    services.api_manager.upload_chunk.assert_called_once()


def test_update_chunk(services: Services):
    chunk = Chunk("test_file.txt", 1, "test_hash")
    data = b"Test Content"
    services.create_chunk(chunk, data)
    services.api_manager.update_chunk.reset_mock()

    chunk.hash = "new_hash"
    data = b"New Content"
    services.update_chunk(chunk, data)

    indexed_chunks = services.db_manager.get_chunks("test_file.txt")
    assert len(indexed_chunks) == 1
    indexed_chunk = indexed_chunks[0]
    assert indexed_chunk.file_path == chunk.file_path
    assert indexed_chunk.order == chunk.order
    assert indexed_chunk.hash == chunk.hash
    services.api_manager.update_chunk.assert_called_once()


def test_delete_chunk(services: Services):
    chunk = Chunk("test_file.txt", 1, "test_hash")
    data = b"Test Content"
    services.create_chunk(chunk, data)
    services.api_manager.upload_chunk.reset_mock()

    services.delete_chunk(chunk)

    indexed_chunks = services.db_manager.get_chunks("test_file.txt")
    assert len(indexed_chunks) == 0
    services.api_manager.delete_chunk.assert_called_once()


def test_create_file(services: Services):
    file_path = os.path.join(INDEXED_FOLDER, "test_file.txt")
    content = b"Test Content"
    write_to_file(file_path, content)
    services.create_file(file_path)

    indexed_file = services.db_manager.get_file_by_path("test_file.txt")
    assert indexed_file.file_path == "test_file.txt"
    assert indexed_file.hash == hashlib.sha256(content).hexdigest()
    services.api_manager.upload_file.assert_called_once()
    indexed_chunks = services.db_manager.get_chunks("test_file.txt")
    assert len(indexed_chunks) == math.ceil(len(content) / CHUNK_SIZE)


def test_update_file_create_chunk(services: Services):
    file_path = os.path.join(INDEXED_FOLDER, "test_file.txt")
    content = bytes(generate_random_string(CHUNK_SIZE), 'utf-8')
    write_to_file(file_path, content)
    services.create_file(file_path)
    services.api_manager.update_file.reset_mock()
    services.api_manager.upload_chunk.reset_mock()

    append_to_file(file_path, content)
    services.update_file(file_path)

    indexed_file = services.db_manager.get_file_by_path("test_file.txt")

    assert indexed_file.file_path == "test_file.txt"
    assert indexed_file.hash == hashlib.sha256(content + content).hexdigest()
    services.api_manager.update_file.assert_called_once()
    assert services.api_manager.upload_chunk.call_count == 1
    indexed_chunks = services.db_manager.get_chunks("test_file.txt")
    assert len(indexed_chunks) == 2


def test_update_file_delete_chunk(services: Services):
    file_path = os.path.join(INDEXED_FOLDER, "test_file.txt")
    content = bytes(generate_random_string(CHUNK_SIZE), 'utf-8')
    write_to_file(file_path, content)
    services.create_file(file_path)
    services.api_manager.update_file.reset_mock()
    services.api_manager.upload_chunk.reset_mock()

    write_to_file(file_path, b"")
    services.update_file(file_path)

    indexed_file = services.db_manager.get_file_by_path("test_file.txt")

    assert indexed_file.file_path == "test_file.txt"
    assert indexed_file.hash == hashlib.sha256(b"").hexdigest()
    services.api_manager.update_file.assert_called_once()
    services.api_manager.delete_chunk.assert_called_once()
    indexed_chunks = services.db_manager.get_chunks("test_file.txt")
    assert len(indexed_chunks) == 0


def test_update_file_update_chunk(services: Services):
    file_path = os.path.join(INDEXED_FOLDER, "test_file.txt")
    content = bytes(generate_random_string(CHUNK_SIZE), 'utf-8')
    write_to_file(file_path, content)
    services.create_file(file_path)
    services.api_manager.update_file.reset_mock()
    services.api_manager.upload_chunk.reset_mock()

    append_to_file(file_path, content)
    services.update_file(file_path)

    indexed_file = services.db_manager.get_file_by_path("test_file.txt")

    assert indexed_file.file_path == "test_file.txt"
    assert indexed_file.hash == hashlib.sha256(content + content).hexdigest()
    services.api_manager.update_file.assert_called_once()
    services.api_manager.upload_chunk.assert_called_once()
    indexed_chunks = services.db_manager.get_chunks("test_file.txt")
    assert len(indexed_chunks) == 2


def test_delete_file(services: Services):
    file_path = os.path.join(INDEXED_FOLDER, "test_file.txt")
    content = b"Test Content"
    write_to_file(file_path, content)
    services.create_file(file_path)
    services.api_manager.upload_file.reset_mock()
    services.api_manager.upload_chunk.reset_mock()

    services.delete_file(file_path)

    indexed_file = services.db_manager.get_file_by_path("test_file.txt")
    assert indexed_file is None
    services.api_manager.delete_file.assert_called_once()
    indexed_chunks = services.db_manager.get_chunks("test_file.txt")
    assert len(indexed_chunks) == 0


def test_check_for_offline_changes(services: Services):
    file_path = os.path.join(INDEXED_FOLDER, "test_file.txt")
    content = b"Test Content"
    write_to_file(file_path, content)
    services.api_manager.upload_file.reset_mock()
    services.api_manager.upload_chunk.reset_mock()

    services.check_for_offline_changes()

    indexed_file = services.db_manager.get_file_by_path("test_file.txt")
    assert indexed_file.file_path == "test_file.txt"
    assert indexed_file.hash == hashlib.sha256(content).hexdigest()
    services.api_manager.upload_file.assert_called_once()
    indexed_chunks = services.db_manager.get_chunks("test_file.txt")
    assert len(indexed_chunks) == math.ceil(len(content) / CHUNK_SIZE)


def test_check_for_offline_deletions(services: Services):
    file_path = os.path.join(INDEXED_FOLDER, "test_file.txt")
    content = b"Test Content"
    write_to_file(file_path, content)
    services.create_file(file_path)
    services.api_manager.upload_file.reset_mock()
    services.api_manager.upload_chunk.reset_mock()

    os.remove(file_path)
    services.check_for_offline_changes()

    indexed_file = services.db_manager.get_file_by_path("test_file.txt")
    assert indexed_file is None
    services.api_manager.delete_file.assert_called_once()
    indexed_chunks = services.db_manager.get_chunks("test_file.txt")
    files = services.db_manager.get_files()
    assert len(indexed_chunks) == 0
    assert len(files) == 0


def test_check_for_offline_changes_no_changes(services: Services):
    file_path = os.path.join(INDEXED_FOLDER, "test_file.txt")
    content = b"Test Content"
    write_to_file(file_path, content)
    services.create_file(file_path)
    services.check_for_offline_changes()
    files = services.db_manager.get_files()
    assert len(files) == 1
    indexed_file = services.db_manager.get_file_by_path("test_file.txt")
    assert indexed_file is not None



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum


@dataclass
class Chunk:
    file_path: str
    order: int
    hash: str

    @staticmethod
    def get_updated_chunks(
        old_chunks: list[Chunk], new_chunks: list[Chunk]
    ) -> list[tuple[Chunk, ChunkAction]]:
        updated_chunks = []
        if len(new_chunks) > len(old_chunks):
            # File has been appended to
            for chunk in new_chunks[len(old_chunks):]:
                updated_chunks.append((chunk, ChunkAction.CREATE))
        elif len(new_chunks) < len(old_chunks):
            # File has been truncated
            for chunk in old_chunks[len(new_chunks):]:
                updated_chunks.append((chunk, ChunkAction.DELETE))
        if len(old_chunks) > 0:
            for i, chunk in enumerate(new_chunks[: len(old_chunks)]):
                if chunk.hash != old_chunks[i].hash:
                    updated_chunks.append((chunk, ChunkAction.UPDATE))
        return updated_chunks


class ChunkAction(Enum):
    CREATE = 1
    UPDATE = 2
    DELETE = 3



import os

DEFAULT_DB_PATH = os.path.expanduser("~/.local/share/tesseract/tesseract.db")
DEFAULT_INDEX_PATH = os.path.expanduser("~/tesseract")
DEFAULT_API_URL = "http://localhost:8000"
CONFIG_PATH = os.path.expanduser("~/.config/tesseract/config.ini")



import os
from loguru import logger

from tesseract_client.api_manager import APIManager
from tesseract_client.db_manager import DBManager
from tesseract_client.file import File, Chunk, ChunkAction


class Services:
    """Provides business logic for the application"""

    def __init__(
        self,
        api_manager: APIManager,
        db_manager: DBManager,
        indexed_folder: str,
        chunk_size: int,
    ):
        self.api_manager = api_manager
        self.db_manager = db_manager
        self.indexed_folder = indexed_folder
        self.chunk_size = chunk_size
        logger.debug("Services object initialized")

    def create_file(self, file_path: str):
        """Creates a new file in the database and uploads it to the server"""
        logger.info(f"File created: {file_path}")
        file = File.from_local_file(file_path, self.indexed_folder)
        self.db_manager.create_file(file)
        self.api_manager.upload_file(file)
        with open(file_path, "rb") as fd:
            chunks = file.split_into_chunks(fd, self.chunk_size)
        for chunk, data in chunks:
            self.create_chunk(chunk, data)

    def update_file(self, file_path: str):
        """
        Updates an existing file in the database and uploads it to the server.
        """
        file = File.from_local_file(file_path, self.indexed_folder)
        indexed_file = self.db_manager.get_file_by_path(file.file_path)
        # Check if the file was actually modified
        if indexed_file.hash == file.hash:
            return
        logger.info(f"File updated: {file_path}")
        self.db_manager.update_file(file)
        self.api_manager.update_file(file)
        indexed_chunks = self.db_manager.get_chunks(file.file_path)
        with open(file_path, "rb") as fd:
            updated_chunks = file.get_updated_chunks(
                fd,
                indexed_chunks,
                self.chunk_size,
            )
        logger.info(f"Found {len(updated_chunks)} updated chunks")
        for (chunk, data), action in updated_chunks:
            if action == ChunkAction.CREATE:
                self.create_chunk(chunk, data)
            elif action == ChunkAction.DELETE:
                self.delete_chunk(chunk)
            else:
                self.update_chunk(chunk, data)

    def delete_file(self, file_path: str):
        """Deletes a file from the database and the server"""
        path = File.get_relative_path(file_path, self.indexed_folder)
        chunks = self.db_manager.get_chunks(path)
        for chunk in chunks:
            self.delete_chunk(chunk)
        self.db_manager.delete_file(path)
        self.api_manager.delete_file(path)
        logger.info(f"File deleted: {file_path}")

    def create_chunk(self, chunk: Chunk, data: bytes):
        """Creates a new chunk in the database and uploads it to the server"""
        self.db_manager.create_chunk(chunk)
        self.api_manager.upload_chunk(chunk, data)

    def update_chunk(self, chunk: Chunk, data: bytes):
        """
        Updates an existing chunk in the database and uploads it to the server.
        """
        self.db_manager.update_chunk(chunk)
        self.api_manager.update_chunk(chunk, data)

    def delete_chunk(self, chunk: Chunk):
        """Deletes a chunk from the database and the server"""
        self.db_manager.delete_chunk(chunk.file_path, chunk.order)
        self.api_manager.delete_chunk(chunk.file_path, chunk.order)

    def check_for_offline_changes(self):
        """Checks for changes that were made while offline"""
        logger.debug("Checking for offline changes")
        self._check_for_offline_deletions()
        self._check_for_offline_updates(self.indexed_folder)

    def _check_for_offline_deletions(self):
        """Checks for files that were deleted while offline"""
        folder = self.indexed_folder
        db_files = self.db_manager.get_files()
        local_files = []
        for root, _, files in os.walk(folder):
            for file in files:
                file = os.path.join(root, file)
                file = File.get_relative_path(file, folder)
                local_files.append(file)
        for db_file in db_files:
            if db_file.file_path not in local_files:
                logger.info(f"File deleted while offline: {db_file.file_path}")
                self.delete_file(os.path.join(folder, db_file.file_path))

    def _check_for_offline_updates(self, folder: str = None):
        """Checks for files that were created/modified while offline"""
        for file in os.listdir(folder):
            file = os.path.join(folder, file)
            if os.path.isdir(file):
                self._check_for_offline_updates(file)
            else:
                db_file = self.db_manager.get_file_by_path(
                    File.get_relative_path(file, self.indexed_folder)
                )
                if db_file:
                    if db_file.hash != File.get_file_hash(file):
                        logger.info(f"File updated while offline: {file}")
                        self.update_file(file)
                else:
                    logger.info(f"File created while offline: {file}")
                    self.create_file(file)

    def pull(self):
        """Pulls the latest changes from the server"""
        logger.debug("Pulling latest changes from the server")
        files = self.api_manager.get_files()
        for file in files:
            db_file = self.db_manager.get_file_by_path(file.file_path)
            if db_file:
                if db_file.hash != file.hash:
                    logger.info(f"File updated on server: {file.file_path}")
                    self._update_file_from_server(file)
            else:
                logger.info(f"File created on server: {file.file_path}")
                self._create_file_from_server(file)
        self._check_for_server_deletions(files)

    def _create_file_from_server(self, file: File):
        """Creates a file from the server"""
        self.db_manager.create_file(file)
        path = os.path.join(self.indexed_folder, file.file_path)
        File.create_folder_path_if_not_exists(path, True)
        open(path, "w").close()
        chunks = self.api_manager.get_file_chunks(file.file_path)
        for chunk in chunks:
            self.db_manager.create_chunk(chunk)
            data = self.api_manager.download_chunk(chunk.file_path, chunk.order)
            with open(os.path.join(self.indexed_folder, file.file_path), "ab") as fd:
                fd.write(data)

    def _update_file_from_server(self, file: File):
        """Updates a file from the server"""
        server_chunks = self.api_manager.get_file_chunks(file.file_path)
        db_chunks = self.db_manager.get_chunks(file.file_path)

        # Update database
        updated_chunks = Chunk.get_updated_chunks(db_chunks, server_chunks)
        for chunk, action in updated_chunks:
            if action == ChunkAction.CREATE:
                self.db_manager.create_chunk(chunk)
            elif action == ChunkAction.DELETE:
                self.db_manager.delete_chunk(chunk.file_path, chunk.order)
            else:
                self.db_manager.update_chunk(chunk)

        # Update file
        with open(os.path.join(self.indexed_folder, file.file_path), "rb") as fd:
            chunks = file.split_into_chunks(fd, self.chunk_size)
        db_chunk_data = [chunk[1] for chunk in chunks]
        db_chunk_hashes = [chunk.hash for chunk in db_chunks]
        chunk_data = []
        for chunk in server_chunks:
            if chunk.hash not in db_chunk_hashes:
                data = self.api_manager.download_chunk(chunk.file_path, chunk.order)
                chunk_data.append(data)
            else:
                chunk_data.append(db_chunk_data[db_chunk_hashes.index(chunk.hash)])
        with open(os.path.join(self.indexed_folder, file.file_path), "wb") as fd:
            for data in chunk_data:
                fd.write(data)

    def _check_for_server_deletions(self, files: list[File]):
        """Checks for files that were deleted on the server"""
        db_files = self.db_manager.get_files()
        files = [file.file_path for file in files]
        for db_file in db_files:
            if db_file.file_path not in files:
                logger.info(f"File deleted on server: {db_file.file_path}")
                self._delete_file_locally(db_file.file_path)

    def _delete_file_locally(self, file_path: str):
        """Deletes a file locally"""
        chunks = self.db_manager.get_chunks(file_path)
        self.db_manager.delete_file(file_path)
        for chunk in chunks:
            self.db_manager.delete_chunk(chunk.file_path, chunk.order)
        os.remove(os.path.join(self.indexed_folder, file_path))



import os
import shutil
import keyring
import configparser
from pathlib import Path
from loguru import logger

from tesseract_client import CONFIG_PATH


class NoCredentialsError(Exception):
    pass


def create_config_if_not_exists(index_path: str, db_path: str, api_url: str):
    """Create a config file if it doesn't exist"""
    path = Path(CONFIG_PATH)
    if path.exists():
        return
    path.parent.mkdir(parents=True)

    config = {"index_path": index_path, "db_path": db_path, "api_url": api_url}

    config_parser = configparser.ConfigParser()
    config_parser.read(CONFIG_PATH)
    config_parser["CONFIG"] = config
    with open(CONFIG_PATH, "w") as configfile:
        config_parser.write(configfile)
    logger.debug(f"Created config file at {CONFIG_PATH}")


def update_config_file(index_path=None, db_path=None, api_url=None):
    """Update the config file"""
    config_parser = configparser.ConfigParser()
    config_parser.read(CONFIG_PATH)
    if index_path is not None:
        config_parser["CONFIG"]["index_path"] = index_path
        logger.debug(f"Updated index path in config file to {index_path}")
    if db_path is not None:
        config_parser["CONFIG"]["db_path"] = db_path
        logger.debug(f"Updated db path in config file to {db_path}")
    if api_url is not None:
        config_parser["CONFIG"]["api_url"] = api_url
        logger.debug(f"Updated API URL in config file to {api_url}")
    with open(CONFIG_PATH, "w") as configfile:
        config_parser.write(configfile)


def update_config(index_path=None, db_path=None, api_url=None):
    prev_index_path, prev_db_path, _ = get_config()

    # Move files to new locations
    if index_path is not None:
        index_path = os.path.expanduser(index_path)
        if index_path != prev_index_path:
            shutil.move(prev_index_path, index_path)
            logger.debug(f"Moved indexed files from {prev_index_path} to {index_path}")
    if db_path is not None:
        db_path = os.path.expanduser(db_path)
        if db_path != prev_db_path:
            shutil.move(prev_db_path, db_path)
            logger.debug(f"Moved database from {prev_db_path} to {db_path}")

    update_config_file(index_path, db_path, api_url)


def get_config() -> tuple[str, str, str]:
    """Get the index path, db path, and API URL from the config file"""
    config_parser = configparser.ConfigParser()
    config_parser.read(CONFIG_PATH)
    index_path = config_parser["CONFIG"]["index_path"]
    db_path = config_parser["CONFIG"]["db_path"]
    api_url = config_parser["CONFIG"]["api_url"]
    return index_path, db_path, api_url


def store_credentials(username: str, password: str):
    """Store the username and password in the config file and keyring"""
    config = configparser.ConfigParser()
    config.read(CONFIG_PATH)
    config["CREDENTIALS"] = {"username": username}
    with open(CONFIG_PATH, "w") as configfile:
        config.write(configfile)
    try:
        keyring.set_password("tesseract", username, password)
        logger.info(f"Stored credentials for {username}")
    except keyring.errors.PasswordSetError:
        logger.error("Failed to store credentials")


def get_credentials() -> tuple[str, str]:
    """Get the username and password from the config file and keyring"""
    try:
        config = configparser.ConfigParser()
        config.read(CONFIG_PATH)
        username = config["CREDENTIALS"]["username"]
        password = keyring.get_password("tesseract", username)
        return username, password
    except KeyError:
        raise NoCredentialsError


def delete_credentials():
    """Delete the username and password from the config file and keyring"""
    try:
        config = configparser.ConfigParser()
        config.read(CONFIG_PATH)
        username = config["CREDENTIALS"]["username"]
        config.remove_section("CREDENTIALS")
        with open(CONFIG_PATH, "w") as configfile:
            config.write(configfile)
        keyring.delete_password("tesseract", username)
        logger.debug(f"Deleted credentials for {username}")
    except KeyError:
        raise NoCredentialsError


def clean_up():
    """Delete the indexed folder and database"""
    index_path, db_path, _ = get_config()

    shutil.rmtree(index_path)
    logger.debug(f"Deleted indexed folder {index_path}")

    os.remove(db_path)
    logger.debug(f"Deleted database {db_path}")



import pytest
from itertools import product
from responses import RequestsMock
from tesseract_client.api_manager import (
    APIManager,
    InvalidCredentials,
    ServerError,
    NotFoundError,
    API_URL
)
from tesseract_client.file import File
from tesseract_client.chunk import Chunk


@pytest.fixture
def api_manager():
    api_urls = API_URL("https://example.com")
    return APIManager("user", "pass", api_urls)


@pytest.fixture
def mock():
    with RequestsMock() as rsps:
        yield rsps


def test_signup_successful(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.POST, api_manager.api_urls.signup, status=200)
    api_manager.signup()


def test_signup_username_taken(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.POST, api_manager.api_urls.signup, status=400)

    with pytest.raises(InvalidCredentials):
        api_manager.signup()


def test_signup_server_error(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.POST, api_manager.api_urls.signup, status=500)

    with pytest.raises(ServerError):
        api_manager.signup()


def test_login_successful(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.POST, api_manager.api_urls.login, json={"access_token": "token123"}, status=200)
    api_manager.login()

    assert api_manager.access_token == "token123"


def test_login_invalid_credentials(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.POST, api_manager.api_urls.login, status=401)

    with pytest.raises(InvalidCredentials):
        api_manager.login()


def test_login_server_error(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.POST, api_manager.api_urls.login, status=500)

    with pytest.raises(ServerError):
        api_manager.login()


def test_login_not_found_error(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.POST, api_manager.api_urls.login, status=404)

    with pytest.raises(NotFoundError):
        api_manager.login()


def test_get_chunk_size(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.GET, api_manager.api_urls.get_chunk_size, body="1024", status=200)
    assert api_manager.get_chunk_size() == 1024


def test_upload_file(api_manager: APIManager, mock: RequestsMock):
    file = File("file.txt", "hash123")
    mock.add(mock.POST, api_manager.api_urls.upload_file, status=201)
    api_manager.upload_file(file)


def test_update_file(api_manager: APIManager, mock: RequestsMock):
    file = File("file.txt", "hash123")
    mock.add(mock.PUT, api_manager.api_urls.upload_file, status=200)
    api_manager.update_file(file)


def test_delete_file(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.DELETE, api_manager.api_urls.delete_file, status=204)
    api_manager.delete_file("file.txt")


def test_get_file(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.GET, api_manager.api_urls.get_file, json={"path": "file.txt", "hash": "hash123"}, status=200)
    file = api_manager.get_file("file.txt")

    assert file.file_path == "file.txt"
    assert file.hash == "hash123"


def test_get_files(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.GET, api_manager.api_urls.get_files, json=[{"path": "file.txt", "hash": "hash123"}], status=200)
    files = api_manager.get_files()

    assert len(files) == 1
    assert files[0].file_path == "file.txt"
    assert files[0].hash == "hash123"


def test_get_file_chunks(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.GET, api_manager.api_urls.get_file_chunks, json=[
        {"file_path": "file.txt",
         "order": 0,
         "hash": "hash123"}
    ], status=200)
    chunks = api_manager.get_file_chunks("file.txt")

    assert len(chunks) == 1
    assert chunks[0].file_path == "file.txt"
    assert chunks[0].order == 0
    assert chunks[0].hash == "hash123"


def test_upload_chunk(api_manager: APIManager, mock: RequestsMock):
    chunk = Chunk("file.txt", 0, "hash123")
    mock.add(mock.POST, api_manager.api_urls.upload_chunk, status=201)
    api_manager.upload_chunk(chunk, b"chunk data")


def test_update_chunk(api_manager: APIManager, mock: RequestsMock):
    chunk = Chunk("file.txt", 0, "hash123")
    mock.add(mock.PUT, api_manager.api_urls.upload_chunk, status=200)
    api_manager.update_chunk(chunk, b"chunk data")


def test_delete_chunk(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.DELETE, api_manager.api_urls.delete_chunk, status=204)
    api_manager.delete_chunk("file.txt", 0)


def test_get_chunk_meta(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.GET, api_manager.api_urls.get_chunk, json={
        "file_path": "file.txt",
        "order": 0,
        "hash": "hash123"
    }, status=200)
    chunk = api_manager.get_chunk_meta("file.txt", 0)

    assert chunk.file_path == "file.txt"
    assert chunk.order == 0
    assert chunk.hash == "hash123"


def test_download_chunk(api_manager: APIManager, mock: RequestsMock):
    mock.add(mock.GET, api_manager.api_urls.get_chunk_url, json={"url": "https://example.com/download"}, status=200)

    mock.add(mock.GET, "https://example.com/download", body=b"chunk data", status=200)
    data = api_manager.download_chunk("file.txt", 0)

    assert data == b"chunk data"


api_methods = [
    (lambda mngr: mngr.api_urls.get_chunk_size,
     lambda mock: mock.GET,
     lambda mngr: mngr.get_chunk_size, ()),

    (lambda mngr: mngr.api_urls.upload_file,
     lambda mock: mock.POST,
     lambda mngr: mngr.upload_file, (File("file.txt", "hash123"),)),

    (lambda mngr: mngr.api_urls.upload_file,
     lambda mock: mock.PUT,
     lambda mngr: mngr.update_file, (File("file.txt", "hash123"),)),

    (lambda mngr: mngr.api_urls.delete_file,
     lambda mock: mock.DELETE,
     lambda mngr: mngr.delete_file, ("file.txt",)),

    (lambda mngr: mngr.api_urls.get_file,
     lambda mock: mock.GET,
     lambda mngr: mngr.get_file, ("file.txt",)),

    (lambda mngr: mngr.api_urls.get_files,
     lambda mock: mock.GET,
     lambda mngr: mngr.get_files, ()),

    (lambda mngr: mngr.api_urls.get_file_chunks,
     lambda mock: mock.GET,
     lambda mngr: mngr.get_file_chunks, ("file.txt",)),

    (lambda mngr: mngr.api_urls.upload_chunk,
     lambda mock: mock.POST,
     lambda mngr: mngr.upload_chunk, (Chunk("file.txt", 0, "hash123"), b"chunk data")),

    (lambda mngr: mngr.api_urls.upload_chunk,
     lambda mock: mock.PUT,
     lambda mngr: mngr.update_chunk, (Chunk("file.txt", 0, "hash123"), b"chunk data")),

    (lambda mngr: mngr.api_urls.delete_chunk,
     lambda mock: mock.DELETE,
     lambda mngr: mngr.delete_chunk, ("file.txt", 0)),

    (lambda mngr: mngr.api_urls.get_chunk,
     lambda mock: mock.GET,
     lambda mngr: mngr.get_chunk_meta, ("file.txt", 0)),

    (lambda mngr: mngr.api_urls.get_chunk_url,
     lambda mock: mock.GET,
     lambda mngr: mngr.download_chunk, ("file.txt", 0)),
]


@pytest.mark.parametrize("url,method,function,args", api_methods)
def test_expired_token_renewed(
    api_manager: APIManager,
    mock: RequestsMock,
    url: callable,
    method: callable,
    function: callable,
    args: tuple
):
    mock.add(method(mock), url(api_manager), status=401)
    mock.add(mock.POST, api_manager.api_urls.login, json={"access_token": "token1234"}, status=200)

    with pytest.raises(InvalidCredentials):
        function(api_manager)(*args)

    assert api_manager.access_token == "token1234"


error_scenarios = [
    (404, NotFoundError),
    (500, ServerError),
]


@pytest.mark.parametrize("method_details,error_scenario", list(product(api_methods, error_scenarios)))
def test_api_methods_errors(api_manager: APIManager, mock: RequestsMock, method_details, error_scenario):
    url, method, function, args = method_details
    status_code, expected_exception = error_scenario

    mock.add(method(mock), url(api_manager), status=status_code)
    with pytest.raises(expected_exception):
        function(api_manager)(*args)



import sqlite3
from loguru import logger

from tesseract_client.file import File, Chunk


class DBManager:
    def __init__(self, db_path: str):
        self.db_path = db_path
        logger.debug(f"DBManager initialized with database path: {db_path}")

    def __enter__(self):
        self.db = sqlite3.connect(self.db_path, check_same_thread=False)
        self.cursor = self.db.cursor()
        self.init_db()
        logger.debug("Database connection established")
        return self

    def __exit__(self, *_):
        self.db.close()
        logger.debug("Database connection closed")

    def init_db(self):
        """Initializes the database tables if they don't already exist"""
        try:
            self.cursor.execute(
                """
                CREATE TABLE IF NOT EXISTS files (
                    file_path TEXT PRIMARY KEY,
                    hash TEXT NOT NULL
                )
            """
            )
            self.cursor.execute(
                """
                CREATE TABLE IF NOT EXISTS chunks (
                    file_path TEXT NOT NULL,
                    order_num INTEGER NOT NULL,
                    chunk_hash TEXT NOT NULL,
                    PRIMARY KEY (file_path, order_num),
                    FOREIGN KEY (file_path) REFERENCES files (file_path)
                )
            """
            )
            self.db.commit()
            logger.debug("Initialized database tables")
        except sqlite3.Error as e:
            logger.error(f"Error initializing database tables: {e}")
            raise

    def create_file(self, file: File):
        """Creates a file object in the database"""
        try:
            self.cursor.execute(
                "INSERT INTO files (file_path, hash) VALUES (?, ?)",
                (file.file_path, file.hash),
            )
            self.db.commit()
            logger.debug(f"File created in database: {file.file_path}")
        except sqlite3.Error as e:
            logger.error(f"Error creating file in database: {e}")
            raise

    def create_chunk(self, chunk: Chunk):
        """Creates a chunk object in the database"""
        try:
            self.cursor.execute(
                "INSERT INTO chunks (file_path, order_num, chunk_hash) VALUES (?, ?, ?)",
                (chunk.file_path, chunk.order, chunk.hash),
            )
            self.db.commit()
            logger.debug(f"Chunk created in database: {chunk.file_path} {chunk.order}")
        except sqlite3.Error as e:
            logger.error(f"Error creating chunk in database: {e}")
            raise

    def update_file(self, file: File):
        """Updates a file object in the database"""
        try:
            self.cursor.execute(
                "UPDATE files SET hash=? WHERE file_path=?", (file.hash, file.file_path)
            )
            self.db.commit()
            logger.debug(f"File updated in database: {file.file_path}")
        except sqlite3.Error as e:
            logger.error(f"Error updating file in database: {e}")
            raise

    def update_chunk(self, chunk: Chunk):
        """Updates a chunk object in the database"""
        try:
            self.cursor.execute(
                "UPDATE chunks SET chunk_hash=? WHERE file_path=? AND order_num=?",
                (chunk.hash, chunk.file_path, chunk.order),
            )
            self.db.commit()
            logger.debug(f"Chunk updated in database: {chunk.file_path} {chunk.order}")
        except sqlite3.Error as e:
            logger.error(f"Error updating chunk in database: {e}")
            raise

    def delete_file(self, file_path: str):
        """Deletes a file object from the database"""
        try:
            self.cursor.execute("DELETE FROM files WHERE file_path=?", (file_path,))
            self.db.commit()
            logger.debug(f"File deleted from database: {file_path}")
        except sqlite3.Error as e:
            logger.error(f"Error deleting file from database: {e}")
            raise

    def delete_chunk(self, file_path: str, order: int):
        """Deletes a chunk object from the database"""
        try:
            self.cursor.execute(
                "DELETE FROM chunks WHERE file_path=? AND order_num=?",
                (file_path, order),
            )
            self.db.commit()
            logger.debug(f"Chunk deleted from database: {file_path} {order}")
        except sqlite3.Error as e:
            logger.error(f"Error deleting chunk from database: {e}")
            raise

    def get_chunks(self, file_path: str):
        """Gets all chunks for a file from the database"""
        try:
            self.cursor.execute(
                "SELECT file_path, order_num, chunk_hash FROM chunks WHERE file_path=?",
                (file_path,),
            )
            chunks = [
                Chunk(file_path=row[0], order=row[1], hash=row[2])
                for row in self.cursor.fetchall()
            ]
            logger.debug(f"Fetched chunks from database for file path: {file_path}")
            return chunks
        except sqlite3.Error as e:
            logger.error(f"Database error while fetching chunks: {e}")
            raise

    def get_files(self):
        """Gets all files from the database"""
        try:
            self.cursor.execute("SELECT file_path, hash FROM files")
            files = [
                File(file_path=row[0], hash=row[1]) for row in self.cursor.fetchall()
            ]
            logger.debug("Fetched files from database")
            return files
        except sqlite3.Error as e:
            logger.error(f"Database error while fetching files: {e}")
            raise

    def get_file_by_path(self, file_path: str):
        """Gets a file from the database by file path"""
        try:
            self.cursor.execute(
                "SELECT file_path, hash FROM files WHERE file_path=?", (file_path,)
            )
            row = self.cursor.fetchone()
            if row is None:
                return None
            file = File(file_path=row[0], hash=row[1])
            logger.debug(f"Fetched file from database by path: {file_path}")
            return file
        except sqlite3.Error as e:
            logger.error(f"Database error while fetching file by path: {e}")
            raise



from tesseract_client.chunk import Chunk, ChunkAction


def test_get_updated_chunks():
    indexed_chunks = [
        Chunk(
            file_path="test.txt",
            order=1,
            hash="hash1"
        ),
        Chunk(
            file_path="test.txt",
            order=2,
            hash="hash2"
        ),
        Chunk(
            file_path="test.txt",
            order=3,
            hash="hash3"
        )
    ]
    chunks = indexed_chunks
    updated_chunks = Chunk.get_updated_chunks(indexed_chunks, chunks)
    assert len(updated_chunks) == 0

    chunks = [
        Chunk(
            file_path="test.txt",
            order=1,
            hash="hash1"
        ),
        Chunk(
            file_path="test.txt",
            order=2,
            hash="hash2"
        ),
        Chunk(
            file_path="test.txt",
            order=3,
            hash="hash4"
        )
    ]
    updated_chunks = Chunk.get_updated_chunks(indexed_chunks, chunks)
    assert len(updated_chunks) == 1
    assert updated_chunks[0][0].order == 3
    assert updated_chunks[0][1] == ChunkAction.UPDATE

    chunks = [
        Chunk(
            file_path="test.txt",
            order=1,
            hash="hash1"
        ),
        Chunk(
            file_path="test.txt",
            order=2,
            hash="hash4"
        ),
        Chunk(
            file_path="test.txt",
            order=3,
            hash="hash3"
        )
    ]
    updated_chunks = Chunk.get_updated_chunks(indexed_chunks, chunks)
    assert len(updated_chunks) == 1
    assert updated_chunks[0][0].order == 2
    assert updated_chunks[0][1] == ChunkAction.UPDATE

    chunks = [
        Chunk(
            file_path="test.txt",
            order=1,
            hash="hash2"
        ),
        Chunk(
            file_path="test.txt",
            order=2,
            hash="hash3"
        ),
    ]
    updated_chunks = Chunk.get_updated_chunks(indexed_chunks, chunks)
    assert len(updated_chunks) == 3
    assert updated_chunks[0][0].order == 3
    assert updated_chunks[0][1] == ChunkAction.DELETE
    assert updated_chunks[1][0].order == 1
    assert updated_chunks[1][1] == ChunkAction.UPDATE
    assert updated_chunks[2][0].order == 2
    assert updated_chunks[2][1] == ChunkAction.UPDATE
